#!/usr/bin/env python
import dbus
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

import math

import pygtk
import gtk
import gobject

import sys
import time

from util import *
from videosink import VideoSink

from telepathy.interfaces import (
    CONNECTION_INTERFACE_CONTACTS,
    CONNECTION_INTERFACE_ALIASING)

try:
    import pygst
    pygst.require('0.10')

    import gst
except ImportError, e:
    raise SystemExit("Gst-Python couldn't be found! (%s)" % (e[0]))

from telepathy.interfaces import (
    CONN_INTERFACE
    )

class CallChannelUI(gtk.Window):
    WIDTH=280
    HEIGHT=-1
    def __init__ (self, bus, channel=None, request=None):
        assert channel or request
        gtk.Window.__init__(self)
        self.connect('destroy', self.window_closed)

        self.member_store = gtk.ListStore (
            gobject.TYPE_STRING, gobject.TYPE_UINT)

        info_frame = gtk.Frame('Call Info')
        info_frame.set_size_request(self.WIDTH, self.HEIGHT)
        vbox = gtk.VBox()
        info_frame.add(vbox)

        self.table = gtk.Table(4, 2)
        self.table.set_col_spacings(3)
        self.table.set_border_width(5)

        self.state_label = gtk.Label()
        self.flags_label = gtk.Label()
        self.reason_label = gtk.Label()
        self.details_label = gtk.Label()

        self.outputs = {}
        self.previews = []

        labels = ["State", "Flags", "Reason", "Details" ]
        for x in xrange(len(labels)):
            l = gtk.Label(labels[x] + ':')
            l.set_justify(gtk.JUSTIFY_RIGHT)
            l.set_alignment(1,0)

            self.table.attach(l, 0, 1, x, x +1,
                xoptions = gtk.FILL, yoptions = 0)

        labels = [self.state_label, self.flags_label,
            self.reason_label, self.details_label]
        for x in xrange(len(labels)):
            labels[x].set_justify(gtk.JUSTIFY_FILL)
            labels[x].set_alignment(0,0)
            self.table.attach(labels[x], 1, 2, x, x +1, yoptions = 0)

        vbox.pack_start(self.table, False, False, 3)

        t = self.treeview = gtk.TreeView (self.member_store)
        t.set_headers_visible (True)
        t.set_rules_hint (True)
        t.insert_column_with_attributes (-1, "Members",
            gtk.CellRendererText(), text=0)
        t.insert_column_with_data_func (-1, "Flags",
            gtk.CellRendererText(), self.render_member_flags)

        scrolled = gtk.ScrolledWindow ()
        scrolled.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.add (t)
        vbox.pack_start(scrolled, True, True, 3)

        vbox = gtk.VBox(False, 3)
        vbox.pack_start(info_frame, True, True, 3)

        button = gtk.Button ("Accept")
        button.connect('clicked', lambda x: self.channel.accept())
        vbox.pack_start(button, False)

        button = gtk.Button ("Hangup")
        button.connect('clicked',
            lambda x: self.channel.hangup(
                CALL_STATE_CHANGE_REASON_REQUESTED,"", "Hangup clicked"))
        vbox.pack_start(button, False)

        self.send_audio = gtk.ToggleButton ("Send Audio")
        self.send_audio.connect('clicked',
            lambda x: self.set_sending(0, x.get_active()))
        vbox.pack_start(self.send_audio, False)

        self.send_video = gtk.ToggleButton ("Send Video")
        self.send_video.connect('clicked',
            lambda x: self.set_sending(1, x.get_active()))
        vbox.pack_start(self.send_video, False)

        self.hbox = gtk.HBox(False, 3)
        self.add(self.hbox)

        self.output_table = gtk.Table (1,1)
        self.output_table_filled = {}

        self.hbox.pack_start(self.output_table, True, True, 3)
        self.hbox.pack_end(vbox, False, False, 3)

        if channel != None:
            self.set_channel(channel)
            self.channel.ringing()
        else:
            request.req.connect_to_signal("Failed", self.req_failed)
            self.state_label.set_text("Pending Channel")

        print "Showing window"
        self.show_all()

    def req_failed(self, error, message):
        self.state_label.set_text("Failed: %s (%s)"%  (error, message))

    def set_channel(self, channel):
        self.channel = channel
        channel.set_call_state_callback (self.update_channel_state)
        self.update_channel_state(channel)

        channel.content_added_cb = self.new_content
        channel.content_removed_cb = self.content_removed
        channel.members_updated_cb = self.members_updated_cb

        for c in self.channel.contents:
            self.new_content(c)

    def update_channel_state(self, channel):
        self.state_label.set_text(call_state_to_s(channel.state))
        self.flags_label.set_text(call_flags_to_s(channel.flags))
        self.reason_label.set_text("actor: %d, %d -- %s" % channel.reason)
        self.details_label.set_text(str(dict(channel.details)))

    def new_content(self, content):
        content.stream_added_cb = self.new_stream
        content.stream_removed_cb = self.remove_stream

        if content.mtype == 1:
            frame = gtk.Frame("Local Preview")
            frame.set_border_width(3)

            vsink = VideoSink(self.channel.conference.pipeline,
                sync = False, async = True)
            frame.add(vsink)
            self.add_frame_to_table (frame)
            frame.show_all()

            self.channel.conference.pipeline.add(vsink.sink)
            vsink.sink.set_state(gst.STATE_PLAYING)

            src_pad = content.session.source.get_request_pad("src%d")
            sink_pad = vsink.sink.get_request_pad("sink%d")
            src_pad.link(sink_pad)

            self.previews.append(vsink)

        for s in content.streams:
            self.new_stream(s, content)

    def new_stream(self, stream, content):
        properties = content.obj.GetAll (CALL_CONTENT,
            dbus_interface=dbus.PROPERTIES_IFACE)
        self_handle = self.channel.conn.Get (CONN_INTERFACE, 'SelfHandle',
            dbus_interface=dbus.PROPERTIES_IFACE)

        if content.mtype == 1: # Telepathy media type video
            participant = stream.fsstream.participant
            stream.set_src_pad_added_callback (
                lambda stream, fsstream, pad, codec: \
                self.new_video_output(participant, stream, pad, codec))
            self.add_output(participant.cname)

        stream.set_senders_changed_callback (
            lambda stream, senders, updated, removed: \
            self.senders_changed_cb(stream, senders,
                stream.fsstream.participant.cname))

    def remove_stream (self, stream, content):
        if content.mtype == 1: # Telepathy media type video
            participant = stream.fsstream.participant
            self.remove_output(participant.cname)

    def senders_changed_cb(self, stream, senders, name):
        self_handle = self.channel.conn.Get (CONN_INTERFACE, 'SelfHandle',
            dbus_interface=dbus.PROPERTIES_IFACE)
        for k, v in senders.iteritems():
            if k == self_handle:
                if v == 1:
                    pass
                elif stream.fsstream.session.type == 0:
                    self.send_audio.set_active(v != 0)
                else:
                    self.send_video.set_active(v != 0)
            elif stream.fsstream.session.type == 1: # Video
                if v != 0 and self.outputs.has_key(name):
                    self.outputs[name].show()
                elif v == 0 and self.outputs.has_key(name):
                    self.outputs[name].hide()

    def add_frame_to_table (self, frame):
        position = 1
        while position in self.output_table_filled.values():
            position += 1

        print "Adding frame to position " + str (position)

        self.output_table_filled[frame] = position

        rows = self.output_table.get_property("n-rows")
        columns = self.output_table.get_property("n-columns")

        # Prefer expanding colums, but keep mostly squared
        if position > (rows * columns):
            if (rows < columns):
                rows += 1
            else:
                columns += 1
        self.output_table.resize (rows, columns)

        # Work out where to add ourselves
        mside = int (math.ceil(math.sqrt(position)))

        # Last position
        last = (mside ** 2)

        # Column filing first
        if position <= (last - mside):
            row = position - ((mside - 1)  ** 2)
            column = mside
        else:
            row = mside
            column = mside + position - last

        self.output_table.attach (frame, column - 1, column,
            row - 1, row, xoptions = 0, yoptions = 0)

    def add_output(self, name):
        gtk.gdk.threads_enter()

        assert not self.outputs.has_key(name)

        frame = gtk.Frame(name)
        frame.set_border_width(3)
        label = gtk.Label("Waiting for video stream...")
        label.set_size_request(320, 240)
        frame.add(label)

        self.outputs[name] = frame
        self.add_frame_to_table (frame)

        frame.show_all()

        gtk.gdk.threads_leave()

    def remove_output(self, name):
        frame = self.outputs.pop(name)
        vsink = frame.get_child()
        if vsink != None and type(vsink) == VideoSink:
            vsink.sink.set_state(gst.STATE_NULL)
            self.channel.conference.pipeline.remove(vsink.sink)

        # Frame might have already been removed when parent was removed
        if frame.parent != None:
            self.output_table.remove(frame)
            position = self.output_table_filled.pop(frame)
            mposition = max(self.output_table_filled.values())

            if position > mposition:
                print "Could make the table smaller"

            self.resize(1,1)

    def new_video_output(self, participant, stream, pad, codec):
        gtk.gdk.threads_enter()

        frame = self.outputs[participant.cname]

        child = frame.get_child()

        if type(child)== gtk.Label:
            frame.remove(frame.get_child())
            vsink = VideoSink(self.channel.conference.pipeline)
            frame.add(vsink)
            frame.show_all()

            self.channel.conference.pipeline.add(vsink.sink)
            vsink.sink.set_state(gst.STATE_PLAYING)
        else:
            vsink = child

        sinkpad = vsink.sink.get_request_pad("sink%d")
        stream.add_sink(sinkpad, vsink.sink)

        pad.link(sinkpad)

        gtk.gdk.threads_leave()

    def window_closed(self, snake):
        self.channel.hangup(CALL_STATE_CHANGE_REASON_REQUESTED,
            "", "User closed the window")
        self.channel.close()

        for o in self.outputs.values():
            if type(o.get_child()) == gtk.Label:
                continue

            vsink = o.get_child()
            vsink.sink.set_state (gst.STATE_NULL)
            self.channel.conference.pipeline.remove(vsink.sink)

        for v in self.previews:
            v.sink.set_state (gst.STATE_NULL)
            self.channel.conference.pipeline.remove(v.sink)

    def set_sending(self, media_type, sending = True):
        found = False

        for c in self.channel.contents:
            if not c.session.type == media_type:
                continue

            found = True

            for s in c.streams:
                s.set_sending(sending)

        if not found and sending:
            self.channel.add_content('new_content', media_type)

    def content_removed_idle(self):
        for k in self.outputs.keys():
            self.remove_output(k)

    def content_removed(self, content):
        for c in self.channel.contents:
            if c.session.type == content.session.type:
                break
        else:
            if content.session.type == 0:
                self.send_audio.set_active(False)
            else:
                self.send_video.set_active(False)
                gobject.idle_add(self.content_removed_idle)

    def members_updated_cb (self):
        handles =  self.channel.members.keys()
        attributes = self.channel.conn.GetContactAttributes (
            dbus.Array(handles, signature="u"),
            dbus.Array([CONNECTION_INTERFACE_ALIASING], signature="s"),
            True,
            dbus_interface = CONNECTION_INTERFACE_CONTACTS)

        self.member_store.clear()
        for handle, flag in self.channel.members.iteritems():
            props = attributes[handle]
            alias = props[CONNECTION_INTERFACE_ALIASING + "/alias"]
            self.member_store.append ((alias, flag))

    def render_member_flags (self, column, cell, model, iter):
        (flags, ) = model.get (iter, 1)
        f = []
        if flags & 1:
            f.append ("Ringing")
        if flags & 2:
            f.append ("Held")

        cell.set_property("text", ", ".join(f))
