import threading

try:
    import pygtk
    pygtk.require("2.0")

    import gtk, gtk.glade, gobject, gtk.gdk
    import gobject
except ImportError, e:
    raise SystemExit("PyGTK couldn't be found ! (%s)" % (e[0]))

try:
    import pygst
    pygst.require('0.10')

    import gst
except ImportError, e:
    raise SystemExit("Gst-Python couldn't be found! (%s)" % (e[0]))

class VideoSinkBin(gst.Bin):
    def __init__ (self, name, pipeline, sync, async):
        gst.Bin.__init__(self, name)
        self.vsink = None
        self.sync = sync
        self.async = async

        self.connect ("element-added", self.element_added_cb)

        self.funnel = fsfunnel = gst.element_factory_make("fsfunnel")
        colorspace = gst.element_factory_make("ffmpegcolorspace")
        videoscale = gst.element_factory_make("videoscale")
        vsink = gst.element_factory_make("gconfvideosink")

        self.add(fsfunnel)
        self.add(colorspace)
        self.add(videoscale)
        self.add(vsink)

        fsfunnel.link(colorspace)
        colorspace.link(videoscale)
        videoscale.link(vsink)

        self.exposedc = threading.Condition()
        self.exposed = False


    def element_added_cb (self, bin, element):
        if isinstance(element, gst.Bin):
            element.connect ("element-added", self.element_added_cb)
            for e in element.elements():
                self.element_added_cb (element, e)

        if isinstance(element, gst.interfaces.XOverlay):
            self.vsink = element
            self.vsink.set_property("sync", self.sync)
            self.vsink.set_property("async", self.async)
            try:
                self.vsink.set_property("force-aspect-ratio", True)
            except Exception:
                pass

    def expose_event(self, *args):
        if not self.exposed:
            self.exposedc.acquire()
            self.exposed = True
            self.exposedc.notify()
            self.exposedc.release()

    def sync_handler(self, bus, message, window):
        if message.src == self.vsink and \
            message.type == gst.MESSAGE_ELEMENT and \
            message.structure.has_name("prepare-xwindow-id"):

            self.exposedc.acquire()
            while not self.exposed:
                self.exposedc.wait()
            gtk.gdk.threads_enter()
            self.vsink.set_xwindow_id (window.xid)
            self.vsink.expose()
            gtk.gdk.threads_leave()
            self.exposedc.release()

            return gst.BUS_DROP

        return gst.BUS_PASS

    def get_request_pad(self, name):
        pad = self.funnel.get_request_pad (name)
        pad.set_active (True)
        ghost = gst.GhostPad(pad.get_name(), pad)
        ghost.set_active (True)
        self.add_pad (ghost)
        return ghost


class VideoSink(gtk.DrawingArea):
    def __init__(self, pipeline, sync = True, async = False):
        gtk.DrawingArea.__init__(self)
        self.set_double_buffered(False)
        self.set_size_request (320, 240)
        self.sink = sink = VideoSinkBin(str(self), pipeline, sync, async)

        self.connect ("expose-event", sink.expose_event)

        bus = pipeline.get_bus()
        bus.enable_sync_message_emission()
        bus.connect("sync-message",
            lambda b,m: sink.sync_handler (b, m, self.window))

